@@ -27,13 +27,16 @@ from __future__ import annotations
 try:
     from urllib.parse import urljoin
 except ImportError:  # pragma: no cover
-    from urlparse import urljoin
+    from urlparse import urljoin  # pyright: ignore[reportMissingImports]
+
+from typing import Any
 
 import httpx
 import requests
 from httpx import Response as AsyncResponse
 from requests import Response
 from requests.adapters import HTTPAdapter
+from requests_toolbelt import MultipartEncoder
 
 from .exceptions import KeycloakConnectionError
 
@@ -59,17 +62,20 @@ class ConnectionManager:
     :type cert: Union[str,Tuple[str,str]]
     :param max_retries: The total number of times to retry HTTP requests.
     :type max_retries: int
+    :param pool_maxsize: The maximum number of connections to save in the pool.
+    :type pool_maxsize: int
     """
 
     def __init__(
         self,
         base_url: str,
         headers: dict | None = None,
-        timeout: int = 60,
-        verify: bool = True,
+        timeout: int | None = 60,
+        verify: bool | str = True,
         proxies: dict | None = None,
         cert: str | tuple | None = None,
         max_retries: int = 1,
+        pool_maxsize: int | None = None,
     ) -> None:
         """
         Init method.
@@ -91,21 +97,33 @@ class ConnectionManager:
         :type cert: Union[str,Tuple[str,str]]
         :param max_retries: The total number of times to retry HTTP requests.
         :type max_retries: int
+        :param pool_maxsize: The maximum number of connections to save in the pool.
+        :type pool_maxsize: int
         """
         self.base_url = base_url
         self.headers = headers
         self.timeout = timeout
         self.verify = verify
+        self.proxies = proxies
         self.cert = cert
+        self.max_retries = max_retries
+        self.pool_maxsize = pool_maxsize
         self._s = requests.Session()
         self._s.auth = lambda x: x  # don't let requests add auth headers
 
         # retry once to reset connection with Keycloak after  tomcat's ConnectionTimeout
         # see https://github.com/marcospereirampj/python-keycloak/issues/36
         for protocol in ("https://", "http://"):
-            adapter = HTTPAdapter(max_retries=max_retries)
+            adapter_kwargs = {"max_retries": max_retries}
+            if pool_maxsize is not None:
+                adapter_kwargs["pool_maxsize"] = pool_maxsize
+            adapter = HTTPAdapter(**adapter_kwargs)  # pyright: ignore[reportArgumentType]
             # adds POST to retry whitelist
-            allowed_methods = set(adapter.max_retries.allowed_methods)
+            allowed_methods = (
+                set(adapter.max_retries.allowed_methods)
+                if adapter.max_retries.allowed_methods
+                else set()
+            )
             allowed_methods.add("POST")
             adapter.max_retries.allowed_methods = frozenset(allowed_methods)
 
@@ -114,9 +132,16 @@ class ConnectionManager:
         if proxies:
             self._s.proxies.update(proxies)
 
-        self.async_s = httpx.AsyncClient(verify=verify, mounts=proxies, cert=cert)
-        self.async_s.auth = None  # don't let requests add auth headers
-        self.async_s.transport = httpx.AsyncHTTPTransport(retries=1)
+        self.async_s = httpx.AsyncClient(
+            verify=verify,
+            mounts=proxies,
+            cert=cert,
+            limits=httpx.Limits(
+                max_connections=100 if pool_maxsize is None else pool_maxsize,
+                max_keepalive_connections=20,
+            ),
+        )
+        self.async_s.auth = None  # pyright: ignore[reportAttributeAccessIssue]
 
     async def aclose(self) -> None:
         """Close the async connection on delete."""
@@ -129,7 +154,7 @@ class ConnectionManager:
             self._s.close()
 
     @property
-    def base_url(self) -> str:
+    def base_url(self) -> str | None:
         """
         Return base url in use for requests to the server.
 
@@ -139,11 +164,11 @@ class ConnectionManager:
         return self._base_url
 
     @base_url.setter
-    def base_url(self, value: str) -> None:
+    def base_url(self, value: str | None) -> None:
         self._base_url = value
 
     @property
-    def timeout(self) -> int:
+    def timeout(self) -> int | None:
         """
         Return timeout in use for request to the server.
 
@@ -153,11 +178,11 @@ class ConnectionManager:
         return self._timeout
 
     @timeout.setter
-    def timeout(self, value: int) -> None:
+    def timeout(self, value: int | None) -> None:
         self._timeout = value
 
     @property
-    def verify(self) -> bool:
+    def verify(self) -> bool | str:
         """
         Return verify in use for request to the server.
 
@@ -167,11 +192,25 @@ class ConnectionManager:
         return self._verify
 
     @verify.setter
-    def verify(self, value: bool) -> None:
+    def verify(self, value: bool | str) -> None:
         self._verify = value
 
     @property
-    def cert(self) -> str | tuple:
+    def proxies(self) -> dict | None:
+        """
+        Return proxies in use for request to the server.
+
+        :returns: Proxies
+        :rtype: dict | None
+        """
+        return self._proxies
+
+    @proxies.setter
+    def proxies(self, value: dict | None) -> None:
+        self._proxies = value
+
+    @property
+    def cert(self) -> str | tuple | None:
         """
         Return client certificates in use for request to the server.
 
@@ -181,11 +220,39 @@ class ConnectionManager:
         return self._cert
 
     @cert.setter
-    def cert(self, value: str | tuple) -> None:
+    def cert(self, value: str | tuple | None) -> None:
         self._cert = value
 
     @property
-    def headers(self) -> dict:
+    def max_retries(self) -> int:
+        """
+        Return maximum number of retries in use for requests to the server.
+
+        :returns: Maximum number of retries
+        :rtype: int
+        """
+        return self._max_retries
+
+    @max_retries.setter
+    def max_retries(self, value: int) -> None:
+        self._max_retries = value
+
+    @property
+    def pool_maxsize(self) -> int | None:
+        """
+        Return the maximum number of connections to save in the pool.
+
+        :returns: Pool maxsize
+        :rtype: int or None
+        """
+        return self._pool_maxsize
+
+    @pool_maxsize.setter
+    def pool_maxsize(self, value: int | None) -> None:
+        self._pool_maxsize = value
+
+    @property
+    def headers(self) -> dict | None:
         """
         Return header request to the server.
 
@@ -195,7 +262,7 @@ class ConnectionManager:
         return self._headers
 
     @headers.setter
-    def headers(self, value: dict) -> None:
+    def headers(self, value: dict | None) -> None:
         self._headers = value or {}
 
     def param_headers(self, key: str) -> str | None:
@@ -207,7 +274,7 @@ class ConnectionManager:
         :returns: If the header parameters exist, return its value.
         :rtype: str
         """
-        return self.headers.get(key)
+        return (self.headers or {}).get(key)
 
     def clean_headers(self) -> None:
         """Clear header parameters."""
@@ -233,6 +300,9 @@ class ConnectionManager:
         :param value: Value to be added.
         :type value: str
         """
+        if self.headers is None:
+            self.headers = {}
+
         self.headers[key] = value
 
     def del_param_headers(self, key: str) -> None:
@@ -242,9 +312,12 @@ class ConnectionManager:
         :param key: Key of the header parameters.
         :type key: str
         """
+        if self.headers is None:
+            return
+
         self.headers.pop(key, None)
 
-    def raw_get(self, path: str, **kwargs: dict) -> Response:
+    def raw_get(self, path: str, **kwargs: Any) -> Response:  # noqa: ANN401
         """
         Submit get request to the path.
 
@@ -256,6 +329,9 @@ class ConnectionManager:
         :rtype: Response
         :raises KeycloakConnectionError: HttpError Can't connect to server.
         """
+        if self.base_url is None:
+            msg = "Unable to perform GET call with base_url missing."
+            raise AttributeError(msg)
         try:
             return self._s.get(
                 urljoin(self.base_url, path),
@@ -269,20 +345,23 @@ class ConnectionManager:
             msg = "Can't connect to server"
             raise KeycloakConnectionError(msg) from e
 
-    def raw_post(self, path: str, data: dict, **kwargs: dict) -> Response:
+    def raw_post(self, path: str, data: dict | str | MultipartEncoder, **kwargs: Any) -> Response:  # noqa: ANN401
         """
         Submit post request to the path.
 
         :param path: Path for request.
         :type path: str
         :param data: Payload for request.
-        :type data: dict
+        :type data: dict | str | MultipartEncoder
         :param kwargs: Additional arguments
         :type kwargs: dict
         :returns: Response the request.
         :rtype: Response
         :raises KeycloakConnectionError: HttpError Can't connect to server.
         """
+        if self.base_url is None:
+            msg = "Unable to perform POST call with base_url missing."
+            raise AttributeError(msg)
         try:
             return self._s.post(
                 urljoin(self.base_url, path),
@@ -297,20 +376,24 @@ class ConnectionManager:
             msg = "Can't connect to server"
             raise KeycloakConnectionError(msg) from e
 
-    def raw_put(self, path: str, data: dict, **kwargs: dict) -> Response:
+    def raw_put(self, path: str, data: dict | str | MultipartEncoder, **kwargs: Any) -> Response:  # noqa: ANN401
         """
         Submit put request to the path.
 
         :param path: Path for request.
         :type path: str
         :param data: Payload for request.
-        :type data: dict
+        :type data: dict | str | MultipartEncoder
         :param kwargs: Additional arguments
         :type kwargs: dict
         :returns: Response the request.
         :rtype: Response
         :raises KeycloakConnectionError: HttpError Can't connect to server.
         """
+        if self.base_url is None:
+            msg = "Unable to perform PUT call with base_url missing."
+            raise AttributeError(msg)
+
         try:
             return self._s.put(
                 urljoin(self.base_url, path),
@@ -325,7 +408,7 @@ class ConnectionManager:
             msg = "Can't connect to server"
             raise KeycloakConnectionError(msg) from e
 
-    def raw_delete(self, path: str, data: dict | None = None, **kwargs: dict) -> Response:
+    def raw_delete(self, path: str, data: dict | None = None, **kwargs: Any) -> Response:  # noqa: ANN401
         """
         Submit delete request to the path.
 
@@ -339,6 +422,10 @@ class ConnectionManager:
         :rtype: Response
         :raises KeycloakConnectionError: HttpError Can't connect to server.
         """
+        if self.base_url is None:
+            msg = "Unable to perform DELETE call with base_url missing."
+            raise AttributeError(msg)
+
         try:
             return self._s.delete(
                 urljoin(self.base_url, path),
@@ -353,7 +440,7 @@ class ConnectionManager:
             msg = "Can't connect to server"
             raise KeycloakConnectionError(msg) from e
 
-    async def a_raw_get(self, path: str, **kwargs: dict) -> AsyncResponse:
+    async def a_raw_get(self, path: str, **kwargs: Any) -> AsyncResponse:  # noqa: ANN401
         """
         Submit get request to the path.
 
@@ -365,6 +452,10 @@ class ConnectionManager:
         :rtype: Response
         :raises KeycloakConnectionError: HttpError Can't connect to server.
         """
+        if self.base_url is None:
+            msg = "Unable to perform GET call with base_url missing."
+            raise AttributeError(msg)
+
         try:
             return await self.async_s.get(
                 urljoin(self.base_url, path),
@@ -376,20 +467,29 @@ class ConnectionManager:
             msg = "Can't connect to server"
             raise KeycloakConnectionError(msg) from e
 
-    async def a_raw_post(self, path: str, data: dict, **kwargs: dict) -> AsyncResponse:
+    async def a_raw_post(
+        self,
+        path: str,
+        data: dict | str | MultipartEncoder,
+        **kwargs: Any,  # noqa: ANN401
+    ) -> AsyncResponse:
         """
         Submit post request to the path.
 
         :param path: Path for request.
         :type path: str
         :param data: Payload for request.
-        :type data: dict
+        :type data: dict | str | MultipartEncoder
         :param kwargs: Additional arguments
         :type kwargs: dict
         :returns: Response the request.
         :rtype: Response
         :raises KeycloakConnectionError: HttpError Can't connect to server.
         """
+        if self.base_url is None:
+            msg = "Unable to perform POST call with base_url missing."
+            raise AttributeError(msg)
+
         try:
             return await self.async_s.request(
                 method="POST",
@@ -403,20 +503,29 @@ class ConnectionManager:
             msg = "Can't connect to server"
             raise KeycloakConnectionError(msg) from e
 
-    async def a_raw_put(self, path: str, data: dict, **kwargs: dict) -> AsyncResponse:
+    async def a_raw_put(
+        self,
+        path: str,
+        data: dict | str | MultipartEncoder,
+        **kwargs: Any,  # noqa: ANN401
+    ) -> AsyncResponse:
         """
         Submit put request to the path.
 
         :param path: Path for request.
         :type path: str
         :param data: Payload for request.
-        :type data: dict
+        :type data: dict | str | MultipartEncoder
         :param kwargs: Additional arguments
         :type kwargs: dict
         :returns: Response the request.
         :rtype: Response
         :raises KeycloakConnectionError: HttpError Can't connect to server.
         """
+        if self.base_url is None:
+            msg = "Unable to perform PUT call with base_url missing."
+            raise AttributeError(msg)
+
         try:
             return await self.async_s.put(
                 urljoin(self.base_url, path),
@@ -433,7 +542,7 @@ class ConnectionManager:
         self,
         path: str,
         data: dict | None = None,
-        **kwargs: dict,
+        **kwargs: Any,  # noqa: ANN401
     ) -> AsyncResponse:
         """
         Submit delete request to the path.
@@ -448,6 +557,10 @@ class ConnectionManager:
         :rtype: Response
         :raises KeycloakConnectionError: HttpError Can't connect to server.
         """
+        if self.base_url is None:
+            msg = "Unable to perform DELETE call with base_url missing."
+            raise AttributeError(msg)
+
         try:
             return await self.async_s.request(
                 method="DELETE",
@@ -462,20 +575,24 @@ class ConnectionManager:
             raise KeycloakConnectionError(msg) from e
 
     @staticmethod
-    def _prepare_httpx_request_content(data: dict | str | None) -> dict:
+    def _prepare_httpx_request_content(data: dict | str | None | MultipartEncoder) -> dict:
         """
         Create the correct request content kwarg to `httpx.AsyncClient.request()`.
 
         See https://www.python-httpx.org/compatibility/#request-content
 
         :param data: the request content
-        :type data: dict | str | None
+        :type data: dict | str | None | MultipartEncoder
         :returns: A dict mapping the correct kwarg to the request content
         :rtype: dict
         """
+        if isinstance(data, MultipartEncoder):
+            return {"content": data.to_string()}
+
         if isinstance(data, str):
             # Note: this could also accept bytes, Iterable[bytes], or AsyncIterable[bytes]
             return {"content": data}
+
         return {"data": data}
 
     @staticmethod
