@@ -524,7 +524,7 @@ class KeycloakOpenID:
 
         return res
 
-    def userinfo(self, token: str) -> dict:
+    def userinfo(self, token: str) -> dict | bytes:
         """
         Get the user info object.
 
@@ -536,7 +536,7 @@ class KeycloakOpenID:
         :param token: Access token
         :type token: str
         :returns: Userinfo object
-        :rtype: dict
+        :rtype: dict | bytes
         """
         orig_bearer = (self.connection.headers or {}).get("Authorization")
         self.connection.add_param_headers("Authorization", "Bearer " + token)
@@ -548,9 +548,9 @@ class KeycloakOpenID:
             else self.connection.del_param_headers("Authorization")
         )
         res = raise_error_from_response(data_raw, KeycloakGetError)
-        if not isinstance(res, dict):
+        if not isinstance(res, (dict, bytes)):
             msg = (
-                "Unexpected response type from userinfo. Expected 'dict', "
+                "Unexpected response type from userinfo. Expected 'dict | bytes', "
                 f"received '{type(res)}', value: '{res}'."
             )
             raise TypeError(msg)
@@ -902,7 +902,12 @@ class KeycloakOpenID:
 
         return list(set(permissions))
 
-    def uma_permissions(self, token: str, permissions: str = "", **extra_payload: Any) -> list:  # noqa: ANN401
+    def uma_permissions(
+        self,
+        token: str,
+        permissions: str | list | dict | tuple | set = "",
+        **extra_payload: Any,  # noqa: ANN401
+    ) -> list:
         """
         Get UMA permissions by user token with requested permissions.
 
@@ -956,7 +961,9 @@ class KeycloakOpenID:
 
         return res
 
-    def has_uma_access(self, token: str, permissions: str) -> AuthStatus:
+    def has_uma_access(
+        self, token: str, permissions: str | list | dict | tuple | set
+    ) -> AuthStatus:
         """
         Determine whether user has uma permissions with specified user token.
 
@@ -1407,7 +1414,7 @@ class KeycloakOpenID:
 
         return res
 
-    async def a_userinfo(self, token: str) -> dict:
+    async def a_userinfo(self, token: str) -> dict | bytes:
         """
         Get the user info object asynchronously.
 
@@ -1419,7 +1426,7 @@ class KeycloakOpenID:
         :param token: Access token
         :type token: str
         :returns: Userinfo object
-        :rtype: dict
+        :rtype: dict | bytes
         """
         orig_bearer = (self.connection.headers or {}).get("Authorization")
         self.connection.add_param_headers("Authorization", "Bearer " + token)
@@ -1431,9 +1438,9 @@ class KeycloakOpenID:
             else self.connection.del_param_headers("Authorization")
         )
         res = raise_error_from_response(data_raw, KeycloakGetError)
-        if not isinstance(res, dict):
+        if not isinstance(res, (dict, bytes)):
             msg = (
-                "Unexpected response type. Expected 'dict', received "
+                "Unexpected response type. Expected 'dict | bytes', received "
                 f"'{type(res)}', value '{res}'."
             )
             raise TypeError(msg)
@@ -1762,7 +1769,7 @@ class KeycloakOpenID:
     async def a_uma_permissions(
         self,
         token: str,
-        permissions: str = "",
+        permissions: str | list | dict | tuple | set = "",
         **extra_payload: Any,  # noqa: ANN401
     ) -> list:
         """
@@ -1818,7 +1825,9 @@ class KeycloakOpenID:
 
         return res
 
-    async def a_has_uma_access(self, token: str, permissions: str) -> AuthStatus:
+    async def a_has_uma_access(
+        self, token: str, permissions: str | list | dict | tuple | set
+    ) -> AuthStatus:
         """
         Determine whether user has uma permissions with specified user token asynchronously.
 
